# -*- coding: utf-8 -*-
"""
Created on Wed Jan 19 12:49:00 2022

@author: dcr
"""

from .. import *
from deprecated import deprecated


import numpy as np
import tkinter
import tkinter.font
from tkinter import *
from tkinter import ttk
    
class FieldInspector:
    def __init__(self, obj, field):
        self.obj = obj
        self.name = field
        
    def getFormat(self):
        return '{}'
    
    def get(self):
        return getattr(self.obj, self.name)
    
    def getFullPath(self):
        return self.obj.getFullPath() + '/' + self.name
        
class Waveform(Logic):
    def __init__(self, parent, name, wires):
        """
        Collects the progress of signals along the time.
        We try to reduce data by avoiding data duplication.

        Parameters
        ----------
        parent : TYPE
            DESCRIPTION.
        name : TYPE
            DESCRIPTION.
        wires : TYPE
            Can be a list of wires or a list of ports.

        Returns
        -------
        None.

        """
        super().__init__(parent, name)
        
        # wires contain all the elements of the input list, there might be repeatitions
        self.wires = wires if isinstance(wires, list) else [wires]
        self.format = []
        self.data = {}  # this is a dictionary of obj (wire or port) -> data, list of wire values 

        self.uniqueWires = []
        
        for x in self.wires:
            
            if isinstance(x, Wire):
                w = x
                if not(x in self.uniqueWires):
                    self.addIn(x.name, x)
                    self.uniqueWires.append(x)
                    self.data[x] = []            
            elif isinstance(x, InPort) or isinstance(x, OutPort):
                w = x.wire
                if (w is None):
                    raise Exception('Wire is null for port', x.getFullPath())
                if not (w in self.uniqueWires):
                    self.addIn(w.name, w)
                    self.uniqueWires.append(w)
                    self.data[w] = []     
            elif isinstance(x, FieldInspector):
                w = None
                if not(x in self.uniqueWires):
                    self.uniqueWires.append(x)
                    self.data[x] = []                                                
            else:
                raise Exception('Unsupported object class to watch: {}'.format(type(x)))
                

            if isinstance(x, FieldInspector):
               self.format.append(x.getFormat()) 
            elif (w.getWidth() == 1):
                self.format.append('')
            else:
                self.format.append('{:X}')

    @staticmethod
    def getwire(x):
        if isinstance(x, Wire):
            w = x
        elif isinstance(x, InPort) or isinstance(x, OutPort):
            w = x.wire
        else:
            raise Exception('Object is not wire {}'.format(x))
            
        return w
    
    def clear(self):
        for key in self.data.keys():
            self.data[key] = []
        
        
    def clock(self):
        # Data is indexed by wire (or field inspector) to avoid repeats
        # For ports, the value used for x will be its wire
        for x in self.uniqueWires:
            if isinstance(x, FieldInspector):
                self.data[x].append(x.get())
            else:
                w = Waveform.getwire(x)
                
                if not(w in self.data):
                    raise Exception('Wire {}/{} not in data list {}'.format(x, w, self.data.keys()))
                    
                self.data[w].append(w.get())
            
    def getDict(self):
        return self.data

    def draw(self):
        import matplotlib.pyplot as plt
        fig, axs = plt.subplots(len(self.wires)+1, sharex=True)
        
        obj = self.wires[0]
        if (isinstance(obj, FieldInspector)):
            w = obj
            ww = -1
        else:
            w = Waveform.getwire(obj)
            ww = w.getWidth()
        
        
        numclocks = len(self.data[w])
        clock = 1 - np.arange(numclocks*2) % 2
        tclock = 0.5 * np.arange(numclocks*2)
        t =  np.arange(numclocks)
        
        axs[0].step(tclock, clock, 'r', linewidth = 2, where='post')
        axs[0].set_ylabel('clock', rotation=0)
        
        for idx, obj in enumerate(self.wires):
            w = Waveform.getwire(obj)

            axs[idx+1].step(t, self.data[w], linewidth=2, where='post')
            axs[idx+1].set_ylabel(w.name, rotation=0)
        
        return fig, axs
    

    
    def draw_wavedrom(self, shortNames=False):
        """
        

        Parameters
        ----------
        shortNames : bool, optional
            Indicates that the names of the wires should be short (instead of the full
            path). The default is False.

        Returns
        -------
        TYPE
            a widget.

        """
        import nbwavedrom as wave
        return wave.draw(self.get_wavedrom(shortNames))
    
    
    
    def get_wavedrom(self, shortNames=False):
        """
        

        Parameters
        ----------
        shortNames : TYPE, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        ret : TYPE
            DESCRIPTION.

        """
        
        signals = [{"name": "clk", 'wave': 'P'}]
        
        for idx, obj in enumerate(self.wires):
            if (isinstance(obj, FieldInspector)):
                w = obj
                ww = -1
            else:
                w = Waveform.getwire(obj)
                ww = w.getWidth()
                
            fmt = self.format[idx]
            wavedata = 'x'
            wavedatadata = []
            
            data = self.data[w]
            last = 'x'
            numclks = len(data)
            for i in range(numclks):
                v = data[i]
                if (v != last):
                    if (ww == 1):
                        wavedata += '{}'.format(v)
                    else:
                        wavedata += '{}'.format(2)
                        wavedatadata.append(fmt.format(v))
                else:
                    wavedata += '.'
                last = v
                    
            wavedata += 'x'
            
            if (shortNames):
                name = obj.name
            else:
                name = obj.getFullPath()
                
            signals.append({'name': name, 'wave':wavedata, 'data':wavedatadata})

        wavedata = 'P'
        for i in range(numclks):
            wavedata += '.'
        wavedata += 'x'

        signals[0]['wave'] = wavedata
        
        ret = {
            "signal": signals,
            "head": {
                "text": self.name,
                "tock": 0,
            }
        }
        
        return ret
    
    def gui(self, root=None, shortNames=False):
        
        window = WaveformWindow(root, self, shortNames)

class WaveformWindow:
    
    def __init__(self, root, wvf, shortNames):
        from py4hw.gui import getResourceIcon 
        if (root is None):
            self.root = tkinter.Tk()
        else:
            self.root = root

        self.waveform = wvf
        self.shortNames = shortNames
            
        self.root.title('Waveform viewer ' + wvf.name)
        
        font = tkinter.font.Font(size=9)
        
        style = ttk.Style()
        style.configure("Treeview", fg="light yellow")
        style.configure("Treeview", rowheight=23) 
        style.configure("Prolepsis.Treeview", font=font)
        
        self.topPane = PanedWindow(self.root, orient=VERTICAL)

        self.toolbar= Frame(self.topPane, bd=1, relief=RAISED)
        
        icon_zo = getResourceIcon('zoomout24.png')
        icon_zi = getResourceIcon('zoomin24.png')
        
        zoomInButton = ttk.Button(self.toolbar,  image=icon_zi, text="Zoom in", command=self.zoomIn)
        zoomOutButton = ttk.Button(self.toolbar, image=icon_zo, text="Zoom out", command=self.zoomOut)

        zoomInButton.pack(side=tkinter.LEFT)
        zoomOutButton.pack(side=tkinter.LEFT)

        self.topPane.add(self.toolbar)

        self.mainPane = PanedWindow(self.topPane, orient=HORIZONTAL)
        self.topPane.add(self.mainPane)
        
        self.hierarchyFrame = Frame(self.mainPane) # PanedWindow(self.mainPane, relief = SUNKEN, width=50, height=100)
        #self.hierarchyPane.pack()
        

        self.hierarchyTree()

        self.mainPane.add(self.hierarchyFrame)

        self.linesPane = PanedWindow(self.mainPane, relief = SUNKEN, width=100, height=100)
        self.mainPane.add(self.linesPane)

        self.hclock = 20
        self.hw = self.hclock * self.getNumClocks()
        
        self.canvas = Canvas(self.linesPane, bg='white', scrollregion=(0,0, self.hw,1000))
        #self.canvas.configure(yscrollcommand=self.scroll.set)
        self.canvas.configure(yscrollincrement='20')
        
        self.h_scroll = ttk.Scrollbar(self.linesPane, orient=HORIZONTAL, command=self.canvas.xview)
        self.v_scroll = ttk.Scrollbar(self.linesPane, orient=VERTICAL, command=self.sync_scrolls)
        self.canvas.configure(xscrollcommand=self.h_scroll.set, yscrollcommand=self.v_scroll.set)

        
        self.linesPane.add(self.canvas)
        self.h_scroll.pack(side=tkinter.BOTTOM, fill=X)
        self.v_scroll.pack(side=tkinter.RIGHT, fill=Y)

        self.drawWaveforms()

        self.topPane.pack(fill=BOTH, expand=True)

        self.root.mainloop()
        
        self.foreground = 'black'
        

    def zoomIn(self):
        self.hclock *= 1.1
        self.hw = self.hclock * self.getNumClocks()

        self.canvas.config(scrollregion=(0, 0, self.hw, 1000))
        self.redraw()
    
    def zoomOut(self):
        self.hclock /= 1.1
        self.hw = self.hclock * self.getNumClocks()

        self.canvas.config(scrollregion=(0, 0, self.hw, 1000))
        self.redraw()
    
    def sync_scrolls(self, *args):
        self.hierarchyTree.yview(*args)
        self.v_scroll.set(*args)
        self.canvas.yview_moveto(args[0])
    
    def yview(self, *args):
        print(*args)
        self.hierarchyTree.yview(*args)
        self.canvas.yview(*args)
        
    def hierarchyTree(self):
    
        tree_scroll = Scrollbar(self.hierarchyFrame)
        tree_scroll.pack(side=RIGHT, fill=Y)

        self.scroll = tree_scroll 
        
        self.hierarchyTree = ttk.Treeview(self.hierarchyFrame, yscrollcommand=tree_scroll.set)
        tv = self.hierarchyTree  
        
        tree_scroll.config(command=self.yview)
        
        self.map_id_obj = {}
    
        #tv.bind('<<TreeviewSelect>>', self.callback)
        
        #self.hierarchyPane.add(tv)        
        #self.hierarchyPane.pack(fill=BOTH, expand=YES)
        tv.pack(side=LEFT, fill=BOTH, expand=YES)
        
        wd = self.waveform.getDict()
        
        for wav in wd.keys():
            if (self.shortNames):
                name = wav.name
            else:
                name = wav.getFullPath()

            dc_iid = tv.insert("", tkinter.END, text=name, open=True)
        
        # self.map_id_obj[dc_iid] = self.sys
    
        # self.populateTree(tv, dc_iid, self.sys)
        
        # print(tv.cget("displaycolumns"))
        #tv.config(columns=[ 'instance', 'type'])
        #tv.config(displaycolumns=[ 0, 1])
        #tv.heading('instance', text='Instance name')
        #tv.heading("type", text="Type")
        tv.heading("#0", text="Name")
        tv.config(selectmode=tkinter.BROWSE)
        tv.tag_configure("ally", background="green")
        #tv.tag_bind("char", "<Double-Button-1>", event)
        
        #tv.config(style="Prolepsis.Treeview")
        tv["style"] = "Prolepsis.Treeview"
        #tv.config(bg="light yellow")
        print(ttk.Style().lookup("Prolepsis.Treeview", "background"))

    def redraw(self):
        self.canvas.delete('all')
        self.drawWaveforms()
        self.canvas.update()

    def getNumClocks(self):
        wd = self.waveform.getDict()
        return len(list(wd.values())[0])
    
    def drawWaveforms(self):
        off = 43
        vspace = 21
        wd = self.waveform.getDict()

        self.setColor('blue')
        
        hclock = self.hclock # this is the width of the clock (in pixels ?)
        vsig = 15
        vtext = 7
        htrans = 3
        
        for idx, obj in enumerate(self.waveform.wires):
            
            if (isinstance(obj, FieldInspector)):
                w = obj
                ww = -1
            else:
                w = Waveform.getwire(obj)
                ww = w.getWidth()
                
            data = wd[w]
            fmt = self.waveform.format[idx]
            
            if (ww == 1):
                lastval = None
            else:
                lastval = '?' # arbitrary value 
            
            for clk in range(len(data)):
                val = data[clk]
                if (ww == 1):
                    # binary wires
                    self.drawLine(clk*hclock, off + idx*vspace - val*vsig, 
                                  (clk+1)*hclock, off + idx*vspace - val*vsig)     
                    if not(lastval is None):
                        # draw the transition at the beginning
                        self.drawLine(clk*hclock, off + idx*vspace - lastval*vsig, 
                                  clk*hclock, off + idx*vspace - val*vsig)     
                    lastval = val 
                else:
                    # non binary wires

                    if not(lastval is None) and (lastval != val):
                        # draw the transition at the beginning
                        self.drawLine(clk*hclock - htrans, off + idx*vspace - vsig, 
                                  clk*hclock + htrans, off + idx*vspace )     
                        self.drawLine(clk*hclock - htrans, off + idx*vspace , 
                                  clk*hclock + htrans, off + idx*vspace - vsig)     
                        self.drawLine(clk*hclock + htrans, off + idx*vspace - vsig, 
                                  (clk+1)*hclock - htrans, off + idx*vspace - vsig)     
                        self.drawLine(clk*hclock + htrans, off + idx*vspace , 
                                  (clk+1)*hclock - htrans, off + idx*vspace )     

                        self.drawText(clk*hclock + htrans, off + idx*vspace - vtext, fmt.format(val), 'w')
                    else:
                        self.drawLine(clk*hclock - htrans, off + idx*vspace - vsig, 
                                  (clk+1)*hclock - htrans, off + idx*vspace - vsig)     
                        self.drawLine(clk*hclock - htrans, off + idx*vspace , 
                                  (clk+1)*hclock - htrans, off + idx*vspace )
                    lastval = val 
        
    def setColor(self, color):
        self.foreground = color
        
    def drawLine(self, x0, y0, x1, y1):
        self.canvas.create_line(x0, y0, x1, y1, fill=self.foreground)

    def drawText(self, x, y, text, anchor):
        if (anchor == 'w'):
            ha = 'left'
        elif (anchor == 'e'):
            ha = 'right'
        elif (anchor == 'c'):
            ha = 'center'
            
        self.canvas.create_text(x,y, anchor=anchor, text=text)
        
@deprecated # use Waveform, just maintened for reference
class OldWaveform(Logic):

    def __init__(self, parent: Logic, name: str, wires):
        """

        Parameters
        ----------
        parent : Logic
              Parent circuit.
        name : str
            Name of the instance.
        wires
              Wire or list of wire to monitor.

        Returns
        -------
        None.

        """
        super().__init__(parent, name)
        self.waves = {}  # A dictionary keyed by fullname , wavem and data
        self.prevs = {}
        self.ck = {"name": "CK", "wave": "P"}

        self.wires = wires # if isinstance(wires, list) else [wires]
        for x in self.wires:
            if (isinstance(x, Wire)):
                self.addIn(x.name, x)
                self.waves[x] = {"name": x.getFullPath(), "wave": "x", "data": []}
                self.prevs[x] = None
            else:
                raise Exception('We only support wires by now')
                
        # Get simulator
        self.sim = parent
        while self.sim.parent != None:
            self.sim = self.sim.parent

        self.sim.getSimulator().addListener(self)

    def simulatorUpdated(self):
        for x in self.wires:
            if self.prevs[x] == x.get():
                self.waves[x]["wave"] += "."
            elif x.getWidth() == 1:
                self.waves[x]["wave"] += str(x.get())
            else:
                self.waves[x]["wave"] += "2"
                self.waves[x]["data"].append(hex(x.get()))

            self.prevs[x] = x.get()

        self.ck["wave"] += "."

    def get_waveform(self, with_ck=True):
        signals = list(self.waves.values())
        for x in signals:
            x["wave"] += "x"

        if with_ck:
            ck = self.ck.copy()
            ck["wave"] += "x"
            signals.insert(0, ck)

        waveform = {
            "signal": signals,
            "head": {
                "text": self.name,
                "tock": 0,
            }
        }

        return waveform
    
    def draw(self):
        import nbwavedrom as wave
        return wave.draw(self.get_waveform())
    

class Scope(Logic):

    def __init__(self, parent: Logic, name: str, wires):
        """
        Parameters
        ----------
        parent : Logic
            Parent circuit.
        name : str
            Name of the instance.
        wires
            Wire or list of wire to monitor.

        Returns
        -------
        None.
        """

        super().__init__(parent, name)
        self.wires = wires if isinstance(wires, list) else [wires]
        for x in self.wires:
            self.addIn(x.name, x)

        # Get simulator
        sim = parent
        while sim.parent != None:
            sim = sim.parent

        sim.getSimulator().addListener(self)

    def simulatorUpdated(self):
        head = f"Scope [{self.name}]:"
        print(head)

        for x in self.wires:
            print(f"{x.name}={x.get()}")

        print("-" * len(head))




class Sequence(Logic):
    """
    A sequence of value
    """

    def __init__(self, parent: Logic, name: str, values: list(), r: Wire):
        super().__init__(parent, name)
        self.r = self.addOut("r", r)

        self.values = values
        self.n = len(values)
        self.i = 0
        
    def clock(self):
        self.r.prepare(self.values[self.i])
        self.i = ( self.i +1 ) % self.n


class RandomValue(Logic):
    def __init__(self, parent, name, r, mean, stddev):
        super().__init__(parent, name)
        
        self.r = self.addOut('r', r)
        
        self.mean = mean
        self.stddev = stddev
        self.state = '{} <- {}+-{}'.format(self.r.get(), self.mean, self.stddev)
        
    def clock(self):
        import numpy as np
        v = int(np.random.normal(self.mean, self.stddev))
        self.r.prepare(v)
        self.state = '{} <- {}+-{}'.format(v, self.mean, self.stddev)
        
    def update_mean(self, v):
        self.mean = float(v)
        self.label1.config(text="Mean: {}".format(self.mean))
        
    def update_stddev(self, v):
        self.stddev = float(v)
        self.label2.config(text="Stddev: {}".format(self.stddev))
        
    def tkinter_gui(self, parent):
        from tkinter import ttk
        import tkinter as tk
        frame = ttk.Frame(parent, padding="10")
        frame.grid(column=2, row=2, sticky=(tk.W, tk.E, tk.N, tk.S))

        maxv = (2 ** self.r.getWidth()) -1
        
        print('maxv : ', maxv)
        
        # Create sliders and labels
        
        self.label1 = ttk.Label(frame, text="Mean: {}".format(self.mean))
        self.label2 = ttk.Label(frame, text="Stddev: {}".format(self.stddev))

        slider1 = ttk.Scale(frame, from_=0, to=maxv, orient="horizontal", command=self.update_mean)
        slider1.set(self.mean)
        slider2 = ttk.Scale(frame, from_=0, to=maxv, orient="horizontal", command=self.update_stddev)
        slider2.set(self.stddev)
        
        

        # Grid layout for sliders and labels
        slider1.grid(column=0, row=0,  pady=10)
        slider2.grid(column=0, row=1,  pady=10)
        self.label1.grid(column=1, row=0, padx=5)
        self.label2.grid(column=1, row=1, padx=5)
        
        # Configure column weights to make the sliders and labels expandable
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=1)
        
        return frame


# -*- coding: utf-8 -*-
"""
Created on Sat Jan 29 16:28:57 2022

@author: dcr
"""

from .base import Wire
from .logic.arithmetic import *
from .logic.bitwise import *
from .logic.simulation import *
from .logic.storage import *
import math


IEEE754_HP_PRECISION = 10
IEEE754_SP_PRECISION = 23
IEEE754_DP_PRECISION = 52

IEEE754_HP_EXPONENT_BITS = 5
IEEE754_SP_EXPONENT_BITS = 8
IEEE754_DP_EXPONENT_BITS = 11

IEEE754_HP_INF_MANTISA = 0x0
IEEE754_SP_INF_MANTISA = 0x0
IEEE754_DP_INF_MANTISA = 0x0

IEEE754_DP_INFNAN_EXPONENT = 0x7FF
IEEE754_DP_EXPONENT_BIAS = 0x3FF

IEEE754_HP_NAN_MANTISA = 0x200
IEEE754_SP_NAN_MANTISA = 0x400000
IEEE754_DP_NAN_MANTISA = 0x8000000000000

IEEE754_HP_SNAN_MANTISA = 0x001
IEEE754_SP_SNAN_MANTISA = 0x000001
IEEE754_DP_SNAN_MANTISA = 0x0000000000001

class LogicHelper:
    """
    Helper class to create logic in a simpler way, 
    basically automatically creating the required output wires
    """
    
    def __init__(self, parent:Logic):
        self.parent = parent
        self.inum = 0   # instance number
        
    def _getNewName(self)->str:
        while (True):
            name = 'i{}'.format(self.inum)
            self.inum += 1
            if not(name in self.parent.children.keys()):
                return name
        
    def hw_abs(self, a:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Abs(self.parent, name, a, r)
        return r

    def hw_add(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = max(wa,wb)+1
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        Add(self.parent, name, a, b, r)
        return r
    
    def hw_and2(self, a:Wire, b:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        And2(self.parent, name, a, b, r)
        return r

    def hw_and3(self, a:Wire, b:Wire, c:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        And(self.parent, name, [a, b, c], r)
        return r

    def hw_and4(self, a:Wire, b:Wire, c:Wire, d:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        And(self.parent, name, [a, b, c, d], r)
        return r

    def hw_and(self, a:list) -> Wire:
        w = a[0].getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        And(self.parent, name, a, r)
        return r
    
    def hw_bit(self, a:Wire, bit:int) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name)
        Bit(self.parent, name, a, bit, r)
        return r

    def hw_bits_msbf(self, a:Wire) -> list:
        w = a.getWidth()
        name = self._getNewName() + '_'
        r = self.parent.wires(name, w, 1)
        BitsMSBF(self.parent, name, a, r)
        return r
        
    def hw_bits_lsbf(self, a:Wire) -> list:
        w = a.getWidth()
        name = self._getNewName() + '_'
        r = self.parent.wires(name, w, 1)
        BitsLSBF(self.parent, name, a, r)
        return r
    
    def hw_buf(self, a:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Buf(self.parent, name, a, r)
        return r
    
    def hw_concatenate_msbf(self, ins:list) -> Wire:
        name = self._getNewName()
        w = 0
        for inw in ins:
            w += inw.getWidth()
            
        r = self.parent.wire(name, w)
        ConcatenateMSBF(self.parent, name, ins, r)
        return r

    def hw_concatenate_lsbf(self, ins:list) -> Wire:
        name = self._getNewName()
        w = 0
        for inw in ins:
            w += inw.getWidth()
            
        r = self.parent.wire(name, w)
        ConcatenateLSBF(self.parent, name, ins, r)
        return r
    
    def hw_constant(self, width:int, v:int) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name, width)
        Constant(self.parent, name, v, r)
        return r
    
    def hw_delay(self, a:Wire, en:Wire, delay:int) -> Wire:
        wa = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, wa)
        if (en is None):
            en = self.hw_constant(1,1)
        DelayLine(self.parent, name, a, en=en, reset=None, r=r, delay=delay)
        return r
    
    def hw_div(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = max(wa,wb)
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        Div(self.parent, name, a, b, r)
        return r    

    def hw_equal(self, a:Wire, b:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name)
        Equal(self.parent, name, a, b, r)
        return r
 
    def hw_equal_constant(self, a:Wire, b:int) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name)
        EqualConstant(self.parent, name, a, b, r)
        return r



    

    

    def hw_mod(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = max(wa,wb)
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        Mod(self.parent, name, a, b, r)
        return r

    
    
    def hw_signed_div(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = max(wa,wb)
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        SignedDiv(self.parent, name, a, b, r)
        return r    

    
    def hw_if(self, cond:Wire, true_option:Wire, false_option:Wire) -> Wire:
        w = true_option.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Mux2(self.parent, name, cond, false_option, true_option, r)
        return r
            
    def hw_not_equal_constant(self, a:Wire, b:int) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name)
        NotEqualConstant(self.parent, name, a, b, r)
        return r
    
    def hw_gt_constant(self, a:Wire, b:int) -> Wire:
        w = a.getWidth()
        gt = self.parent.wire(self._getNewName())
        eq = self.parent.wire(self._getNewName())
        lt = self.parent.wire(self._getNewName())
        k = self.parent.wire(self._getNewName(), w)
        Constant(self.parent, self._getNewName(), b, k)
        Comparator(self.parent, self._getNewName(), a, k, gt, eq, lt)
        return gt
    
    def hw_lt_constant(self, a:Wire, b:int) -> Wire:
        w = a.getWidth()
        gt = self.parent.wire(self._getNewName())
        eq = self.parent.wire(self._getNewName())
        lt = self.parent.wire(self._getNewName())
        k = self.parent.wire(self._getNewName(), w)
        Constant(self.parent, self._getNewName(), b, k)
        Comparator(self.parent, self._getNewName(), a, k, gt, eq, lt)
        return lt
    
    def hw_ge_constant(self, a:Wire, b:int) -> Wire:
        w = a.getWidth()
        gt = self.parent.wire(self._getNewName())
        eq = self.parent.wire(self._getNewName())
        lt = self.parent.wire(self._getNewName())
        ge = self.parent.wire(self._getNewName())
        k = self.parent.wire(self._getNewName(), w)
        Constant(self.parent, self._getNewName(), b, k)
        Comparator(self.parent, self._getNewName(), a, k, gt, eq, lt)
        Or2(self.parent, self._getNewName(), gt, eq, ge)
        return ge
    
    def hw_mul(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = wa+wb
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        Mul(self.parent, name, a, b, r)
        return r
        
    def hw_mux2(self, sel:Wire, s0:Wire, s1:Wire) -> Wire:
        w = s0.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Mux2(self.parent, name, sel, s0, s1, r)
        return r

    def hw_mux(self, sel:Wire, options:list) -> Wire:
        name = self._getNewName()
        w = options[0].getWidth()
        r = self.parent.wire(name, w)
        Mux(self.parent, name, sel, options, r)
        return r

    def hw_neg(self, a:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Neg(self.parent, name, a, r)
        return r

    def hw_not(self, a:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Not(self.parent, name, a, r)
        return r

    
    def hw_or2(self, a:Wire, b:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Or2(self.parent, name, a, b, r)
        return r

    def hw_or(self, a:list) -> Wire:
        w = a[0].getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Or(self.parent, name, a, r)
        return r

    def hw_or_bits(self, a) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name)
        OrBits(self.parent, name, a, r)
        return r

    def hw_range(self, a:Wire, up:int, down:int) -> Wire:
        w = up - down + 1
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Range(self.parent, name, a, up, down, r)
        return r
    
    def hw_repeat(self, a:Wire, n:int) -> Wire:
        w = n * a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Repeat(self.parent, name, a, r)
        return r

    def hw_select_default(self, sels:list, ins:list, default:Wire) -> Wire:
        name = self._getNewName()
        w = max([obj.getWidth() for obj in ins])
        r = self.parent.wire(name, w)        
        SelectDefault(self.parent, name, sels, ins, default, r)
        return r        
        
    def hw_signed_add(self, a:Wire, b:Wire) -> Wire:
        wa = a.getWidth()
        wb = b.getWidth()
        wr = max(wa,wb)
        name = self._getNewName()
        r = self.parent.wire(name, wr)
        Add(self.parent, name, a, b, r)
        return r

        
    def hw_shift_left_constant(self, a:Wire, n:int) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w+n)
        ShiftLeftConstant(self.parent, name, a, n, r)
        return r


    def hw_sign(self, a:Wire) -> Wire:
        w = 1
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Sign(self.parent, name, a, r)
        return r

    def hw_sub(self, a:Wire, b:Wire) -> Wire:
        w = max(a.getWidth(), b.getWidth())
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Sub(self.parent, name, a, b, r)
        return r

    def hw_reg(self, a:Wire, enable=None, reset=None, reset_value=None) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Reg(self.parent, name, d=a, q=r, enable=enable, reset=reset, reset_value=reset_value)
        return r

    def hw_xor2(self, a:Wire, b:Wire) -> Wire:
        w = a.getWidth()
        name = self._getNewName()
        r = self.parent.wire(name, w)
        Xor2(self.parent, name, a, b, r)
        return r
    
    def sign_extend(self, a:Wire, w) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name, w)
        SignExtend(self.parent, name, a, r)
        return r
    
    def zero_extend(self, a:Wire, w) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name, w)
        ZeroExtend(self.parent, name, a, r)
        return r
    
    def sim_sequence(self, width:int, seq:list) -> Wire:
        name = self._getNewName()
        r = self.parent.wire(name, width)
        Sequence(self.parent, name, seq, r)
        return r

class IntegerHelper:
    @staticmethod
    def sign(v):
        if (v < 0):
            return -1
        else:
            return 1
        
    @staticmethod
    def signed_to_c2(v, w):
        mask = (1<<w)-1
        return v & mask

    @staticmethod
    def c2_to_signed(v, w):
        v = v & ((1<<w)-1)
        
        sign = 1 << (w-1)
        if ((v & sign) > 0):
            # negative
            return (v - (1 << w))
        else:
            return v
        
def signExtend(v, w, nw):
    # converts a signed or unsigned value to a c2 representation
    v = (v & ((1<<w)-1)) # first mask all possible additional bits
    s = (v >> (w-1)) & 1
    ns = 0
    if (s == 1):
        ns = (1 << (nw-w)) - 1
    return v | (ns << w)

class FixedPoint:
    def __init__(self, sign_bit, int_bits, frac_bits, v):
        self.sw = sign_bit
        self.iw = int_bits
        self.fw = frac_bits
        self.v = 0
        
        if (type(v) == int):
            self.intToFixedPoint(v)
        elif (type(v) == float):
            self.floatToFixedPoint(v)
        else:
            raise Exception('Unsupported type {}'.format(type(v)))
        
    def getWidths(self):
        return (self.sw, self.iw, self.fw)
        
    def intToFixedPoint(self, v):
        if (v < 0 and self.sw == 0):
            raise Exception('negative values not supported')
            
        w = self.sw + self.iw + self.fw
        
        maxv = 1 << (self.iw -1)
        
        if (v > maxv):
            raise Exception('Value {} greater than max value: {}'.format(v, maxv))
        
        self.v = (v << self.fw) & ((1<<w)-1)
            
    def floatToFixedPoint(self, v):
        if (v < 0 and self.sw == 0):
            raise Exception('negative values not supported')
        
        w = self.sw + self.iw + self.fw
        self.v = int(v * (1 << self.fw)) & ((1<<w)-1)
        
    def add(self, b):
        if (not(isinstance(b, FixedPoint))):
            b = FixedPoint(self.sw, self.iw, self.fw, b)
            
        r = FixedPoint(self.sw, self.iw, self.fw , 0)
        w = self.sw + self.iw + self.fw
        r.v = (self.v + b.v) & ((1<<w)-1)
        return r

    def sub(self, b):
        if (not(isinstance(b, FixedPoint))):
            b = FixedPoint(self.sw, self.iw, self.fw, b)

        r = FixedPoint(self.sw, self.iw, self.fw , 0)
        w = self.sw + self.iw + self.fw
        r.v = (self.v - b.v) & ((1<<w)-1)
        return r

    def mult(self, b):
        if (not(isinstance(b, FixedPoint))):
            oldb = b
            b = FixedPoint(self.sw, self.iw, self.fw, b)
            #print('old b is ', oldb, 'new b is {:08X}'.format( b.v))
    
        r = FixedPoint(self.sw, self.iw, self.fw , 0)
        w = self.sw + self.iw + self.fw
        av = signExtend(self.v, w, w*2)
        bv = signExtend(b.v, w, w*2)
        r.v = ((av * bv)>>self.fw)  & ((1<<w)-1)
        return r

    def toFloatingPoint(self):
        if (((self.v >> (self.iw+self.fw)) & 1) == 1):
            w = self.sw + self.iw + self.fw
            mask = ((1<<w)-1)
            v = ((self.v ^ mask) + 1) & mask
            return -v / (1<<self.fw)
        return self.v / (1<<self.fw)
        
    def copy():
        return FixedPoint(self.sw, self.iw, self.fw , self.v)
    
    @staticmethod
    def fromRawValue(sw, iw, fw, v):
        r = FixedPoint(sw,iw,fw,0)
        r.v = v
        return r
    
    def createConstant(self, parent, name):
        w = self.sw + self.iw + self.fw
        mask = ((1<<w)-1)
        wire = parent.wire(name, w)
        Constant(parent, name, self.v & mask, wire)
        return wire
    
    def dump(self):
        fmt = f'|[:0{self.sw}b]|[:0{self.iw}b]|[:0{self.fw}b]|'
        fmt = fmt.replace('[','{')
        fmt = fmt.replace(']','}')
        s = self.v >> (self.iw + self.fw) & 1
        i = self.v >> (self.fw) & ((1<<self.iw)-1)
        f = self.v & ((1<<self.fw)-1)
        return fmt.format(s,i,f)
        
class FixedPointHelper:
    pass

class FPNum:
    max_prec = 1 << 200
    
    def __init__(self, *args):
        self.inexact = False
        self.infinity = False
        self.nan = False
        
        if (len(args) == 0):
            pass
        elif len(args) == 1:
            # should be a float
            self.convert_float_to_semp(args[0])
        elif len(args) == 2:
            # should be a int + string (sp, dp)
            if (args[1] == 'hp'):
                self.from_ieee754_hp(args[0])
            elif (args[1] == 'sp'):
                self.from_ieee754_sp(args[0])
            elif (args[1] == 'dp'):
                self.from_ieee754_dp(args[0])
            else:
                raise Exception(f'unknown format {args[1]}')
        elif (len(args) == 3):
            # should be a sem
            self.adjust_sem(args[0], args[1], args[2])
        elif (len(args) == 4):
            # should be a sem
            self.set_semp(args[0], args[1], args[2], args[3])
            self.adjust_semp()
        else:
            raise Exception('invalid parameter')
    
    def components(self):
        return self.s, self.e, self.m, self.p
    
    def copy(self):
        r = FPNum()
        r.set_semp(self.s, self.e, self.m, self.p)
        r.infinity = self.infinity
        r.nan = self.nan
        r.inexact = self.inexact
        return r
    
    def isPositiveInfinity(self):
        return self.infinity and self.s == 1 
    
    def isNegativeInfinity(self):
        return self.infinity and self.s == -1 
    
    def to_float(self):
        if (self.infinity):
            if (self.s == 1): return math.inf
            elif (self.s == -1): return -math.inf
            else: raise Exception('no valid sign')
        if (self.nan):
            return math.nan
        
        from decimal import Decimal
        n2 = Decimal(2)
        r = n2 ** Decimal(self.e)
        r = r * Decimal(self.m)
        r = r / Decimal(self.p)
        r = r * Decimal(self.s)
        return float(r)
    
    def reduceExponentPrecision(self, prec):
        mask = (1 << prec) - 1
        e_bias = mask >> 1 
        
        if (self.e < -(e_bias-1)):
            # subnormal
            while (self.e < -(e_bias-1)):
                self.e += 1 
                self.p = self.p << 1
        elif (self.e >= e_mask):
            self.infinity = True
        
    def reducePrecisionWithRounding(self, prec):
        # prec is the number of bits of the precision
        p = 1 << prec
        r = 0
        i = 0
        
        if (p < self.p):
            while (p < self.p):
                r = (self.m & 1) << i | r
                self.p = self.p >> 1
                self.m = self.m >> 1
                i += 1
            self.inexact = True
            
            if (r > (1<<(i-1))):
                self.m += 1
            
    def reducePrecision(self, prec):
        p = 1 << prec
        
        if (p < self.p):
            while (p < self.p):
                self.p = self.p >> 1
                self.m = self.m >> 1
            self.inexact = True
    
    def compare(self, bref):
        if (self.nan or bref.nan): return 0 
        if (self.infinity and bref.infinity and (self.s == bref.s)): return 0
        if (self.infinity and bref.infinity and (self.s != bref.s)): return 1
        if (self.infinity or bref.infinity): 
            if (self.infinity): return self.s
            else: return -bref.s

        a = FPNum(self.s, self.e, self.m, self.p)
        b = FPNum(bref.s, bref.e, bref.m, bref.p)

        # ensure equal exponent         
        if (a.e > b.e): b.increase_exponent(a.e)
        elif (a.e < b.e): a.increase_exponent(b.e)
        
        assert(a.e == b.e)
        
        # ensure equal precision
        if (a.p > b.p): b.increase_precision(a.p)
        elif (a.p < b.p): a.increase_precision(b.p)
        
        assert(a.p == b.p)

        if (a.m == b.m):  abs_cmp  = 0
        elif (a.m > b.m): abs_cmp = 1
        elif (a.m < b.m): abs_cmp = -1
        else: raise Exception()
            
        if (a.s == 1 and b.s == 1): return abs_cmp
        elif (a.s == -1 and b.s == -1): return -abs_cmp
        elif (a.s == -1 and b.s == 1): return -1
        elif (a.s == 1 and b.s == -1): return 1
        else: raise Exception()

    def convert(self, fmt):
        s, e, m, p = self.s, self.e, self.m, self.p
        s = 0 if s > 0 else 1

        # deal with special cases        
        if (self.infinity) or (self.nan):
            if (fmt == 'hp'):
                e = 0x1F
                if (self.nan): 
                    s = 0
                    m = IEEE754_HP_NAN_MANTISA
                else:
                    m = IEEE754_HP_INF_MANTISA
                x = self.pack_ieee754_hp_parts(s, e, m)
            elif (fmt == 'sp'):
                e = 0xFF
                if (self.nan):
                    s = 0
                    m = IEEE754_SP_NAN_MANTISA
                else:
                    m = IEEE754_SP_INF_MANTISA
                x =  self.pack_ieee754_sp_parts(s, e, m)
            elif (fmt == 'dp'):
                e = 0x7FF
                if (self.nan):
                    s = 0
                    m = IEEE754_DP_NAN_MANTISA
                else:
                    m = IEEE754_DP_INF_MANTISA
                x = self.pack_ieee754_dp_parts(s, e, m)

            return x        
        
        if (m==0):
            if   (fmt == 'hp'): return self.pack_ieee754_hp_parts(s, 0, 0)
            elif (fmt == 'sp'): return self.pack_ieee754_sp_parts(s, 0, 0)
            elif (fmt == 'dp'): return self.pack_ieee754_dp_parts(s, 0, 0)
            
        
        if (fmt == 'hp'):
            e_bias = 15
            e_mask = 0x1F
            p_std = 1 << 10
        elif (fmt == 'sp'):
            e_bias = 127
            e_mask = 0xFF
            p_std = 1 << 23
        elif (fmt == 'dp'):
            e_bias = 1023
            e_mask = 0x7FF
            p_std = 1 << 52
        else:
            raise Exception(f'unknown format: {fmt}')
            
        
        if (e < -(e_bias-1)):
            # if exponent is smaller that e_bias-1, then it is a subnormal
            while (e < -(e_bias-1)):
                e += 1 
                p = p << 1
            e = 0
        else:
            if (e == -(e_bias-1)) and (p > m):
                # also a subnormal number
                e = 0
            else: e = e + e_bias   
            
        if (e < 0):
            # very small number
            e = 0
            m = 0            
        elif (e >= e_mask):
            # infinity
            e = e_mask
            m = 0
        else:
            if (e == 0):
                # subnormal numbers no not need further mantisa processing
                pass
            else:
                if (p > m):
                    m = m << 1
                    e -= 1
                if (m >= (p<<1)):
                    p = p << 1
                    e += 1
                    
                if not(m & p): print('ERROR converting', self.components(), 'to', fmt)
                assert(m & p)   # ensure higher bit is active
                m = m ^ p       # eliminate the higher bit
            
            # compute the standard precision
            while (p > p_std):
                p = p >> 1 
                m = m >> 1 
            while (p < p_std):
                p = p << 1 
                m = m << 1
            
        if (fmt == 'hp'):
            x = self.pack_ieee754_hp_parts(s, e, m)
        elif (fmt == 'sp'):
            x =  self.pack_ieee754_sp_parts(s, e, m)
        elif (fmt == 'dp'):
            x = self.pack_ieee754_dp_parts(s, e, m)

        return x            
        
    def increase_exponent(self, ne):
        while (self.e < ne):
            self.e += 1 
            self.p = self.p << 1

    def increase_precision(self, np):
        while (self.p < np):
            self.p = self.p << 1 
            self.m = self.m << 1
            
    def abs(self):
        return FPNum(1, self.e, self.m, self.p)
        
    def neg(self):
        return FPNum(self.s * -1, self.e, self.m, self.p)
        
    
    def sub(self, bref):
        if (self.nan or bref.nan): return FPNum(self.s, self.e, -1, 0) # nan
        if (self.infinity and bref.infinity and ((self.s * bref.s)==1)): return FPNum(self.s, self.e, -1, 0) # nan
        if (self.infinity or bref.infinity): 
            if (self.infinity): return self.copy() # normal
            if (bref.infinity): return FPNum(bref.s * -1, bref.e, 0, 0) # invert the sign of this infinity
            
        b = FPNum(bref.s * -1, bref.e, bref.m, bref.p)
        return self.add(b)
    
    def add(self, bref):
        if (self.nan or bref.nan): return FPNum(self.s, self.e, -1, 0) # nan
        if (self.infinity and bref.infinity and ((self.s * bref.s)==-1)): return FPNum(self.s, self.e, -1, 0) # nan
        if (self.infinity or bref.infinity): 
            if (self.infinity): return self.copy()
            if (bref.infinity): return bref.copy()
        
        a = FPNum(self.s, self.e, self.m, self.p)
        b = FPNum(bref.s, bref.e, bref.m, bref.p)

        # ensure equal exponent         
        if (a.e > b.e): b.increase_exponent(a.e)
        elif (a.e < b.e): a.increase_exponent(b.e)
        
        assert(a.e == b.e)
        
        # ensure equal precision
        if (a.p > b.p): b.increase_precision(a.p)
        elif (a.p < b.p): a.increase_precision(b.p)
        
        assert(a.p == b.p)
        
        if (a.s == 1 and b.s == 1) or (a.s == -1 and b.s == -1):
            s = a.s
            m = a.m + b.m 
        elif (a.s == 1 and b.s == -1):
            if (a.m > b.m): 
                s = a.s 
                m = a.m - b.m
            else:
                s = b.s
                m = b.m - a.m
        elif (a.s == -1 and b.s == 1):
            if (a.m > b.m):
                s = a.s
                m = a.m - b.m
            else:
                s = b.s 
                m = b.m - a.m
                
        return FPNum(s, a.e, m, a.p)
        
    def div(self, b):
        # Zeri is a special case
        if (b.m == 0):
            if (self.m == 0):
                return FPNum(self.s, self.e, -1, 0) # nan
            else:
                r = self.copy()
                r.infinity = True
                r.s = self.s * b.s
                return r
        
        rs = self.s * b.s
        re = self.e - b.e
        
        rm = int((self.m * b.p  * self.max_prec) / b.m)
        rp = self.p * self.max_prec
        
        return FPNum(rs, re, rm, rp)

    def mul(self, b):
        if (self.nan or b.nan): return FPNum(self.s, self.e, -1, 0) # nan
        if (self.infinity or b.infinity): return FPNum(self.s * b.s, self.e, 0, 0) # infinity
            
        rs = self.s * b.s
        re = self.e + b.e
        
        rm = self.m * b.m
        rp = self.p * b.p
        
        return FPNum(rs, re, rm, rp)
    
    def div2(self, n):
        s , e, m, p = self.s, self.e, self.m, self.p << n
        return FPNum(s, e, m, p)
    
    def sqrt(self, iterations=20):
        x1 = FPNum(self.s, self.e >> 1, self.m, self.p)
        
        # zero is an special case
        if (self.m == 0):
            return self.copy()
        
        if (self.s == -1):
            # sqrt of negative numbers is Nan
            r = self.copy()
            r.nan = True
            return r
        
        #print('computing sqrt({})'.format(self.to_float()))
        
        # Newton-Raphson method 
        for i in range(iterations):
            x2 = self.div(x1)      
            #print(self.to_float(), '/', x1.to_float() , '=', x2.to_float())
            x2 = x2.add(x1)
            #print('    +', x1.to_float() , '=', x2.to_float())
            x2 = x2.div2(1)
            er = x1.sub(x2)
            
            # print('x1:', x1.to_float(), 'x2:', x2.to_float(), 'er:', er.to_float(), er.components())
            
            x1 = x2.copy()
        
        return x2
        
    def from_ieee754_hp(self, v):
        s,e,m = self.unpack_ieee754_hp_parts(v)
        s = 1 if s == 0 else -1
        
        if (e == 0x1F):
            self.set_semp(s, e, m, 0) # special cases are signaled with p = 0 
            return
        
        if (e == 0):
            # subnormal numbers
            e = -16
            m = m
        else:
            e = e - 15            
            m = (1 << 10) | m

        p = 1 << 10
        
        self.set_semp(s, e, m, p)
        self.adjust_semp()

    
    def from_ieee754_sp(self, v):
        s,e,m = self.unpack_ieee754_sp_parts(v)
        s = 1 if s == 0 else -1
        
        if (e == 0xFF):
            self.set_semp(s, e, m, 0) # special cases are signaled with p = 0
            return
        
        if (e == 0):
            # subnormal numbers
            e = -126
            m = m
        else:
            e = e - 127            
            m = (1 << 23) | m

        p = 1 << 23
        
        self.set_semp(s, e, m, p)
        self.adjust_semp()

    def from_ieee754_dp(self, v):
        s,e,m = self.unpack_ieee754_dp_parts(v)
        s = 1 if s == 0 else -1
        
        if (e == 0x7FF):
            self.set_semp(s, e, m, 0) # special cases are signaled with p = 0
            return
        
        if (e == 0):
            # subnormal numbers
            e = -1022
            m = m
        else:
            e = e - 1023
            m = (1 << 52) | m
        p = 1 << 52
        
        self.set_semp(s, e, m, p)
        self.adjust_semp()
    
    @staticmethod
    def unpack_ieee754_hp_parts(v):
        s = (v >> 15) & 1
        e = (v >> 10) & 0x1F
        m = (v & ((1<<10)-1))          
        return s, e, m
    
    @staticmethod
    def unpack_ieee754_sp_parts(v):
        s = (v >> 31) & 1
        e = (v >> 23) & 0xFF
        m = (v & ((1<<23)-1))          
        return s, e, m
    
    @staticmethod
    def unpack_ieee754_dp_parts(v):
        s = (v >> 63) & 1
        e = (v >> 52) & 0x7FF
        m = (v & ((1<<52)-1))  
        return s, e, m

    @staticmethod
    def pack_ieee754_hp_parts(s, e, m):
        return ((s & 1) << 15) | ((e & 0x1F) << 10) | (m & ((1<<10)-1))  

    @staticmethod
    def pack_ieee754_sp_parts(s, e, m):
        return ((s & 1) << 31) | ((e & 0xFF) << 23) | (m & ((1<<23)-1))  

    @staticmethod
    def pack_ieee754_dp_parts(s, e, m):
        return ((s & 1) << 63) | ((e & 0x7FF) << 52) | (m & ((1<<52)-1))  

    def convert_float_to_semp(self, v):
        if (math.isinf(v)):
            self.s , self.e, self.m , self.p =  1 if v > 0 else -1, 0, 0, 0
            self.infinity , self.nan = True, False
            return
        elif (math.isnan(v)):
            self.s , self.e, self.m , self.p =  1, 0, -1, 0
            self.infinity , self.nan = False, True
            return
            
        s, e, m, p = 1, 0, v, 1
        
        if (v == 0) and (math.copysign(1, v) == -1): s = -1
        elif (v < 0):
            s, m = -s, -m
            
        self.adjust_sem(s, e, m)
        self.adjust_semp()
        
    def adjust_sem(self, s, e, m):
        # the goal is to have m in the range of 1 to 2
        self.s = s
        self.e = e
        self.m = m
        self.p = 1
        
        n0 = 0
        n1 = 1
        n2 = 2
        
        if (self.m == 0):
            self.m = int(m)
            return 
        
        # determine exponent
        if (self.m >= 2):
            while (self.m >= 2):
                self.m /= n2
                self.e += 1
        elif (self.m < 1):
            while (self.m < 1):
                self.m *= n2
                self.e -= 1
                
        # determine divisor
        while ((self.m - int(self.m)) > 0):
            self.m *= n2
            self.p *= n2
                
        self.m = int(self.m)
        self.p = int(self.p)
        
    def set_semp(self, s, e, m, p):
        self.s = s
        self.e = e
        self.m = m
        self.p = p
        
        if (p == 0):
            if (m == 0): self.infinity = True 
            else:  self.nan = True
            

    def adjust_semp(self):
        # eliminate decimal values from the mantisa
        assert(isinstance(self.m, int))
        
        if (self.p == 0): return # ignore it for special cases

        # reduce the precision 
        while ((self.m & 1)==0) and ((self.p & 1) == 0):
            self.m = self.m >> 1
            self.p = self.p >> 1
            #print('prec --:', math.log2(self.p))

        # adjust the exponent, mantisa has to be between p and 2p
        p2 = self.p << 1
        
        if (self.m >= p2):
            while (self.m >= p2):
                # increase the exponent and increase the precision
                self.p = self.p << 1
                p2 = self.p << 1
                self.e += 1
                #print('e ++:', self.e)

        elif (self.m < self.p):
            while (self.m < self.p):
                self.m = self.m << 1 
                self.e -= 1                
                #print('e --:', self.e, self.m)
                if (self.m == 0):
                    return
        
            

class FloatingPointHelper:

    @staticmethod
    def min(a, b):
        if (a == 0.0 and b == 0.0):
            sa = math.copysign(1, a)
            sb = math.copysign(1, b)
            
            if (sa < sb):
                return a
            else:
                return b
        return min(a, b)

    @staticmethod
    def max(a, b):
        if (a == 0.0 and b == 0.0):
            sa = math.copysign(1, a)
            sb = math.copysign(1, b)
            
            if (sa > sb):
                return a
            else:
                return b
        return max(a, b)
    
    @staticmethod
    def fp_to_parts(v):
        if (math.isinf(v) or math.isnan(v)):
            raise Exception('invalid value {}'.format(v))
        e = 0
        s=0
        m = v
        if (m < 0):
            # if v is negative activate the sign, and change the value
            s = 1
            m = -m
    
        if (m > 0.0):
            while (m >= 2):
                # iterate until v is < 2 to determine the exponent
                m = m / 2
                e += 1
                #print('e:', e, 'v:', v)
            while (m < 1):
                m = m * 2
                e -= 1
                #print('e:', e, 'v:', v)
            
        return s, e, m

    @staticmethod
    def sp_to_fixed_point_parts(v):
        s,e,m = FloatingPointHelper.fp_to_parts(v)

        if (m == 0):
            return 0,0,0
        else:
            re = e
            rm = int(round(m * (1<<23)))

            return s, re, rm
    
    @staticmethod
    def sp_to_ieee754_parts(v):
        """
        Return the parts of the IEEE 754 representation of v

        Parameters
        ----------
        v : TYPE
            DESCRIPTION.

        Returns
        -------
        int
            sign.
        int
            biased representation of exponent (e+127)
        int
            representation of mantissa int((m-1)<<23) 

        """
        if (math.isinf(v)):
            s = 0 if v > 0 else 1
            e = 255
            m = 0
            return s,e,m
        
        if (math.isnan(v)):
            s = 0 
            e = 255
            m = (1<<23)-1
            return s,e,m
        
        s,e,m = FloatingPointHelper.fp_to_parts(v)

        if (m == 0):
            return 0,0,0
        else:
            if (e >= 128):
                return s,255,0  # infinity
                
            if (e <= -127):
                # denormalized values
                div = math.pow(2, (-127-e))
                #print('e',e,'div', div)
                m = m / div
                re = 0
                rm = int(round(m * (1 << 22)))
            else:
                im = int(round((m-1) * (1<<23)))
                if (im >= (1<<23)):
                    e += 1
                    m /= 2
                    
                re = 127 + e
                rm = int(round((m-1) * (1<<23)))
    
            return s, re, rm

    def dp_to_ieee754_parts(v):
        """
        Return the parts of the IEEE 754 representation of v

        Parameters
        ----------
        v : TYPE
            DESCRIPTION.

        Returns
        -------
        int
            sign.
        int
            biased representation of exponent (e+1023)
        int
            representation of mantissa int((m-1)<<52) 

        """
        if (math.isinf(v)):
            s = 0 if v > 0 else 1
            e = 2047
            m = 0
            return s,e,m

        if (math.isnan(v)):
            s = 0 
            e = 2047
            m = (1<<51)-1
            return s,e,m

            
        s,e,m = FloatingPointHelper.fp_to_parts(v)

        if (m == 0):
            v = math.copysign(1, v)
            s = 0 if v > 0 else 1
            return s,0,0
        else:
            if (e >= 1024):
                return s,2047,0  # infinity
                
            if (e <= -1023):
                # denormalized values
                div = math.pow(2, (-1023-e))
                #print('e',e,'div', div)
                m = m / div
                re = 0
                rm = int(round(m * (1 << 51)))
            else:
                im = int(round((m-1) * (1<<52)))
                if (im >= (1<<52)):
                    e += 1
                    m /= 2
                    
                re = 1023 + e
                rm = int(round((m-1) * (1<<52)))
    
            return s, re, rm

    @staticmethod
    def sp_to_ieee754(v):
        s,e,m = FloatingPointHelper.sp_to_ieee754_parts(v)
        
        r = s << 31
        r = r | (e << 23)
        r = r | (m)
        return r
    
    @staticmethod
    def dp_to_ieee754(v):
        s,e,m = FloatingPointHelper.dp_to_ieee754_parts(v)
        
        r = s << 63
        r = r | (e << 52)
        r = r | (m)
        return r

    @staticmethod
    def parts_to_fp(s, e, m):
        return math.pow(-1, s) * math.pow(2, e) * m

    @staticmethod
    def ieee754_parts_to_sp(s, e, m):
        """
        We build a floating point number from is sign/exponent/mantisa representation
        as it is store in the IEEE754 format

        Parameters
        ----------
        s : int
            sign.
        e : int
            exponent.
        m : TYPE
            mantisa.

        Returns
        -------
        float
            floating point number r = (-1)^s * 2^e * m

        """
        
        # zero is a special case
        if (e == 0 and m == 0):
            if (s == 1):
                return -0.0
            else:
                return 0.0
            
        if (e == 0):
            ef = -126
            mf = m / (1<<23)
        else:
            ef = e-127
            mf = ((1 << 23) | m) / (1<<23)
        
        return FloatingPointHelper.parts_to_fp(s , ef , mf)
    
    @staticmethod
    def ieee754_parts_to_dp(s, e, m):
        """
        We build a floating point number from is sign/exponent/mantisa representation
        as it is store in the IEEE754 format

        Parameters
        ----------
        s : int
            sign.
        e : int
            exponent.
        m : TYPE
            mantisa.

        Returns
        -------
        float
            floating point number r = (-1)^s * 2^e * m

        """
        
        # zero is a special case
        if (e == 0 and m == 0):
            if (s == 1):
                return -0.0
            else:
                return 0.0
            
        if (e == 0):
            ef = -1022
            mf = m / (1<<52)
        else:
            ef = e-1023
            mf = ((1 << 52) | m) / (1<<52)
        
        return FloatingPointHelper.parts_to_fp(s , ef , mf)
    
    @staticmethod
    def ieee754_to_sp(v):
        if (v == 0):
            return 0.0

        s, e, m = FloatingPointHelper.unpack_ieee754_sp_parts(v)
                
        if (e == 255):
            if (m == 0):
                return -math.inf if (s == 1) else math.inf
            else:
                return math.nan
        
        return FloatingPointHelper.ieee754_parts_to_sp(s, e, m)

    @staticmethod
    def ieee754_to_dp(v):
        if (v == 0):
            return 0.0

        s,e,m = FloatingPointHelper.unpack_ieee754_dp_parts(v)
        
        if (e == 0x7FF):
            if (m == 0):
                return -math.inf if (s == 1) else math.inf 
            else:
                return math.nan

        return FloatingPointHelper.ieee754_parts_to_dp(s, e, m)

    @staticmethod
    def unpack_ieee754_sp_parts(v):
        s = v >> 31
        e = (v >> 23) & 0xFF
        m = (v & ((1<<23)-1))          
        return s, e, m
    
    @staticmethod
    def unpack_ieee754_dp_parts(v):
        s = v >> 63
        e = (v >> 52) & 0x7FF
        m = (v & ((1<<52)-1))  
        return s, e, m
    
    
    @staticmethod
    def pack_ieee754_sp_parts(s, e, m):
        return ((s & 1) << 31) | ((e & 0xFF) << 23) | (m & ((1<<23)-1))  

    @staticmethod
    def ieee754_sp_neg(v):
        s, e, m = FloatingPointHelper.unpack_ieee754_sp_parts(v)
        s = s ^ 1
        return FloatingPointHelper.pack_ieee754_sp_parts(s, e, m)
        
    @staticmethod
    def ieee754_stored_internally(v):
        s,e,m = FloatingPointHelper.sp_to_ieee754_parts(v)
        iv = FloatingPointHelper.ieee754_parts_to_sp(s, e, m)
        return iv

    @staticmethod
    def zp_bin(v:int, vl:int) -> str:
        """        
        Generates a the binary representation of an integer value in a number
        of bits. Leading zeros are added as needed.
        
        Parameters
        ----------
        v : int
            value.
        vl : int
            length of the output.

        Returns
        -------
        TYPE
            the binary string.

        """
        if (v >= 0):
            str = bin(v)[2:]
        else:
            # bin function does not handle 2's complement,
            # so let's do it for them
            p = -v
            mask = (1 << vl) -1
            r = (mask - p + 1) & mask
            str = bin(r)[2:] 
            
        return ('0' * (vl-len(str))) + str
    
    
class BitManipulation:
    @staticmethod
    def getFirstBit(v:int) -> int:
        # returns the position of the first active bit
        n = 0
        while (True):
            if (v == 0):
                return n
            n += 1
            v = v >> 1
    
    @staticmethod
    def countLeadingZeros(v:int, w:int) -> int:
        return w - BitManipulation.getFirstBit(v)

    @staticmethod
    def getBit(v:int, bit:int) -> int:
        return (v >> bit) & 0x1
    
class CircuitAnalysis:

    @staticmethod
    def getAllPorts(obj:Logic) -> list:
        ret = []
        
        ret.extend(obj.inPorts)
        ret.extend(obj.outPorts)
            
        return ret
    
    def getAllWireNames(obj:Logic) -> list:
        """
        Return the list of wire names

        Parameters
        ----------
        obj : Logic
            DESCRIPTION.

        Returns
        -------
        None.

        """
        return [x for x in obj._wires.keys()]
    
    @staticmethod
    def getAllPortNames(obj:Logic) -> list:
        """
        Get all port names of an object

        Parameters
        ----------
        obj : Logic
            DESCRIPTION.

        Returns
        -------
        list
            DESCRIPTION.

        """
        ret = []
        
        for p in obj.inPorts:
            ret.append(p.name)
            
        for p in obj.outPorts:
            ret.append(p.name)
            
        return ret

    @staticmethod
    def has_clk_enable(obj:Logic) -> bool:
        input_port_names = [p.name for p in obj.inPorts]
        return 'clk_enable' in input_port_names
        
    @staticmethod
    def getAllPortWires(obj:Logic) -> list:
        """
        Returns a list of the wires connected to all the ports of a circuit

        Parameters
        ----------
        obj : Logic
            the object to analyze.

        Returns
        -------
        list
            list with all the wires connected to the circuit ports.

        """
        ret = []
        
        for p in obj.inPorts:
            ret.append(p.wire)
            
        for p in obj.outPorts:
            ret.append(p.wire)
            
        return ret


# -*- coding: utf-8 -*-
"""
Created on Wed Jan 19 12:36:38 2022

@author: dcr
"""
from .. import *
from .relational import *
from deprecated import deprecated
import math





class And2(Logic):
    """
    Binary And
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

    def propagate(self):
        self.r.put(self.a.get() & self.b.get())
        
        
class And(Logic):
    def __init__(self, parent, name:str, ins, r: Wire):
        super().__init__(parent, name)
        lins = []
        r = self.addOut('r', r)
        w = r.getWidth()
        
        for idx, inv in enumerate(ins):
            lins.append(self.addIn('in{}'.format(idx), inv))

        # store inputs/outputs for RTL generation
        self.r = r
        self.ins = lins
        
        num = len(ins)

        if (num == 1):
            Buf(self, 'and1', ins[0], r)
            return
        
        if (num == 2):
            And2(self, 'and2', ins[0], ins[1], r)
            return
            
        # by now we do an inefficient ladder structure, we should
        # do a more fancy logarithmic design
        
        auxin = lins[0]
        auxout = self.wire('and{}'.format(0), w)
        
        for i in range(num-1):
            # print('creating and{} input0: {}'.format(i, auxin.getFullPath()))
            # print('creating and{} input1: {}'.format(i, lins[i+1].getFullPath()))
            # print('creating and{} output: {}'.format(i, auxout.getFullPath()))
            And2(self, 'and{}'.format(i),  auxin, lins[i+1], auxout)
            auxin = auxout
            if (i == num-3):
                auxout = r
            else:
                auxout = self.wire('and{}'.format(i+1), w)
                
                        
                
class Bit(Logic):
    """ 
    Gets one bit from a wire
    """

    def __init__(self, parent: Logic, name: str, a: Wire, bit, r: Wire):
        super().__init__(parent, name)

        self.a = self.addIn('a', a)
        self.r = self.addOut('r', r)

        self.bit = bit

    def propagate(self):
        value = self.a.get()
        newvalue = (value >> self.bit) & 1
        self.r.put(newvalue)


class BitsMSBF(Logic):
    """
    Returns a list of 1 bit wires from a wider wire 
    in least significant bit first order
    """

    def __init__(self, parent: Logic, name: str, a: Wire, bits):
        super().__init__(parent, name)

        self.a = self.addIn('a', a)

        w = a.getWidth()

        self.bits = []
        for i in range(w):
            self.bits.append(self.addOut('b{}'.format(i), bits[i]))

        self.bits.reverse()
        
    def propagate(self):
        w = self.a.getWidth()
        v = self.a.get()

        for i in range(w):
            self.bits[i].put((v >> i) & 1)
            
class BitsLSBF(Logic):
    """
    Returns a list of 1 bit wires from a wider wire 
    in least significant bit first order
    """

    def __init__(self, parent: Logic, name: str, a: Wire, bits):
        super().__init__(parent, name)

        self.a = self.addIn('a', a)

        w = a.getWidth()

        self.bits = []
        for i in range(w):
            self.bits.append(self.addOut('b{}'.format(i), bits[i]))

    def propagate(self):
        w = self.a.getWidth()
        v = self.a.get()

        for i in range(w):
            self.bits[i].put((v >> i) & 1)

    @deprecated
    def fromWire(parent: Logic, name: str, a: Wire):
        bits = parent.wires(name + '_bits', a.getWidth(), 1)
        BitsLSBF(parent, name, a, bits)
        return bits
    
class Buf(Logic):
    def __init__(self, parent, name: str, a: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn('a', a)
        self.r = self.addOut('r', r)

    def propagate(self):
        self.r.put(self.a.get())

    
class BidirBuf(Logic):
    
    def __init__(self, parent, name: str, pin : Wire , pout : Wire, poe : Wire, bidir : BidirWire):
        super().__init__(parent, name)
        
        if (poe.getWidth() != 1):
            raise Exception('BidirBuf only supports 1 bit output enable')
            
        self.bidir = self.addInOut('bidir', bidir)
        self.pin = self.addOut('pin', pin)
        self.pout = self.addIn('pout', pout)
        self.poe = self.addIn('poe', poe)
        
    def propagate(self):
        if (self.poe.get() ==1):
            self.bidir.put(self.pout.get())
        else:
            self.pin.put(self.bidir.get())

class BufEnable(Logic):
    # This is just a way to control set to zero all bits of a signal
    # depending on an enable signal
    def __init__(self, parent, name, a, en, r):
        super().__init__(parent, name)
        
        self.a = self.addIn('a', a)
        self.en = self.addIn('en', en)
        self.r = self.addOut('r', r)
   
        assert(a.getWidth() == r.getWidth())
    
        re = self.wire('re', r.getWidth())
        Repeat(self, 're', en, re)
        And2(self, 'r', a, re, r)
        
    def structureName(self):
        return 'BufEnable{}'.format(self.a.getWidth())
    

class Constant(Logic):
    """
    A constant value
    """

    def __init__(self, parent: Logic, name: str, value: int, r: Wire):
        super().__init__(parent, name)
        assert(isinstance(value, int))
        assert(isinstance(r, Wire))
        self.value = value
        self.r = self.addOut("r", r)

    def propagate(self):
        self.r.put(self.value)
        #print(self.name, '=', self.value)
        
class Demux(Logic):
    def __init__(self, parent, name, a, sel, r):
        super().__init__(parent, name)
        
        self.addIn('a', a)
        self.addIn('sel', sel)
        
        assert(2**sel.getWidth() == len(r))
        
        ss = self.wires('ss', 2**sel.getWidth(), 1)
        Decoder(self, 'decoder', sel, ss)
        
        for idx, ri in enumerate(r):
            self.addOut('r{}'.format(idx), ri)
            
            BufEnable(self, 'en{}'.format(idx), a, ss[idx], ri)
            
class Nand2(Logic):
    """
    Binary Nand
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

        self.mid = self.wire("Mid", a.getWidth())

        And2(self, "And", a, b, self.mid)
        Not(self, "Not", self.mid, r)

class Nor(Logic):
    """
    Binary Nand
    """

    def __init__(self, parent, name: str, ins, r: Wire):
        super().__init__(parent, name)
        
        lins = []
        
        for idx, inv in enumerate(ins):
            lins.append(self.addIn('in{}'.format(idx), inv))

        r = self.addOut("r", r)

        mid = self.wire("Mid", lins[0].getWidth())
        
        # save inputs/outputs for RTL generation
        self.r = r
        self.ins = lins

        Or(self, "Or", lins, mid)
        Not(self, "Not", mid, r)


class Nor2(Logic):
    """
    Binary Nand
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

        self.mid = self.wire("Mid", a.getWidth())

        Or2(self, "Or", a, b, self.mid)
        Not(self, "Not", self.mid, r)

class Not(Logic):
    def __init__(self, parent, name: str, a: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.r = self.addOut("r", r)

    def propagate(self):
        self.r.put(~self.a.get())

    def getSymbol(self, x, y):
        return NotSymbol(self, x, y)


class Or2(Logic):
    """
    Binary And
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

    def propagate(self):
        self.r.put(self.a.get() | self.b.get())

class OrBits(Logic):
    def __init__(self, parent, name:str, a: Wire, r: Wire):
        super().__init__(parent, name)
        
        self.addIn('a', a)
        self.addOut('r', r)
        
        ab = self.wires('ab', a.getWidth(), 1)
        BitsLSBF(self, 'bits', a, ab)
        Or(self, 'or', ab, r)
    
class Or(Logic):
    def __init__(self, parent, name:str, ins, r: Wire):
        super().__init__(parent, name)
        lins = []
        r = self.addOut('r', r)
        w = r.getWidth()
        
        for idx, inv in enumerate(ins):
            lins.append(self.addIn('in{}'.format(idx), inv))

        # save inputs/outputs for RTL generation
        self.r = r
        self.ins = lins
        
        num = len(ins)

        if (num == 1):
            Buf(self, 'or1', ins[0], r)
            return 
        
        if (num == 2):
            Or2(self, 'or2', ins[0], ins[1], r)
            return
                    
        # by now we do an inefficient ladder structure, we should
        # do a more fancy logarithmic design
        
        auxin = lins[0]
        auxout = self.wire('and{}'.format(0), w)
        
        for i in range(num-1):
            # print('creating and{} input0: {}'.format(i, auxin.getFullPath()))
            # print('creating and{} input1: {}'.format(i, lins[i+1].getFullPath()))
            # print('creating and{} output: {}'.format(i, auxout.getFullPath()))
            Or2(self, 'and{}'.format(i),  auxin, lins[i+1], auxout)
            auxin = auxout
            if (i == num-3):
                auxout = r
            else:
                auxout = self.wire('and{}'.format(i+1), w)    


class ShiftLeftConstant(Logic):
    def __init__(self, parent, name: str, a: Wire, n: int, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn('a', a)
        self.r = self.addOut('r', r)
        self.addParameter('n', n)

    def propagate(self):
        self.r.put(self.a.get() << self.getParameterValue('n'))


class ShiftRightConstant(Logic):
    def __init__(self, parent, name: str, a: Wire, n: int, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn('a', a)
        self.r = self.addOut('r', r)
        self.addParameter('n', n)

    def propagate(self):
        self.r.put(self.a.get() >>  self.getParameterValue('n'))

class RotateLeftConstant(Logic):
    def __init__(self, parent, name: str, a: Wire, n: int, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn('a', a)
        self.r = self.addOut('r', r)
        self.n = n

    def propagate(self):
        a = self.a.get()
        n = self.n
        w = self.a.getWidth()
        self.r.put((a << n) | (a >> (w - n)))


class RotateRightConstant(Logic):
    def __init__(self, parent, name: str, a: Wire, n: int, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn('a', a)
        self.r = self.addOut('r', r)
        self.n = n

    def propagate(self):
        a = self.a.get()
        n = self.n
        w = self.a.getWidth()
        self.r.put((a >> n) | (a << (w - n)))
                
class Xor2(Logic):
    """
    Binary Xor
    """

    def __init__(self, parent, name: str, a: Wire, b: Wire, r: Wire):
        super().__init__(parent, name)
        self.a = self.addIn("a", a)
        self.b = self.addIn("b", b)
        self.r = self.addOut("r", r)

        mid = self.wire("Mid", a.getWidth())
        xout = self.wire("XOut", a.getWidth())
        yout = self.wire("YOut", a.getWidth())

        Nand2(self, "NandMid", a, b, mid)
        Nand2(self, "NandX", a, mid, xout)
        Nand2(self, "NandY", b, mid, yout)
        Nand2(self, "NandR", xout, yout, r)


class Xor(Logic):
    def __init__(self, parent, name:str, ins, r: Wire):
        super().__init__(parent, name)
        lins = []
        r = self.addOut('r', r)
        w = r.getWidth()
        
        for idx, inv in enumerate(ins):
            lins.append(self.addIn('in{}'.format(idx), inv))

        num = len(ins)

        if (num == 2):
            Xor2(self, 'and2', ins[0], ins[1], r)
            return
            
        if (num < 3):
            raise Exception('List should be > 2')

        # by now we do an inefficient ladder structure, we should
        # do a more fancy logarithmic design
        
        auxin = lins[0]
        auxout = self.wire('and{}'.format(0), w)
        
        for i in range(num-1):
            # print('creating and{} input0: {}'.format(i, auxin.getFullPath()))
            # print('creating and{} input1: {}'.format(i, lins[i+1].getFullPath()))
            # print('creating and{} output: {}'.format(i, auxout.getFullPath()))
            Xor2(self, 'and{}'.format(i),  auxin, lins[i+1], auxout)
            auxin = auxout
            if (i == num-3):
                auxout = r
            else:
                auxout = self.wire('and{}'.format(i+1), w)
                

class Mux(Logic):
    def __init__(self, parent, name: str, sel: Wire, ins, r: Wire):
        super().__init__(parent, name)
        
        if (sel.getWidth() != int(math.log2(len(ins)))):
            raise Exception('Invalid length sel bits: {} # ins: {}'.format(sel.getWidth(), int(math.log2(len(ins)))))

        sel = self.addIn('sel', sel)
        r = self.addOut('r', r)

        lins = []
        for idx, inv in enumerate(ins):
            lins.append(self.addIn('in{}'.format(idx), inv))
        
        if (sel.getWidth() == 1):
            # two inputs is a trivial case
            Mux2(self, 'mux2', sel, lins[0], lins[1], r)
            return
            
        # bits contain all the bits from the selection wire
        # which will be the base for creating a logarithmic
        # selection tree
        bits = BitsLSBF.fromWire(self, 'sel', sel)


            
        fp, ip = math.modf(math.log2(len(ins)))
        
        if (fp != 0):
            raise Exception('Input wires are not a power of 2')

        w = len(ins) // 2
        auxin = ins
        auxout = self.wires('l{}'.format(0), w, r.getWidth())

        for i in range(len(bits)):
            for k in range(w):
                Mux2(self, 'm{}_{}'.format(i, k), bits[i], auxin[k * 2 + 0], auxin[k * 2 + 1], auxout[k])

            auxin = auxout
            w = w // 2
            if (w == 1):
                auxout = [r]
            else:
                auxout = self.wires('l{}'.format(i + 1), w, r.getWidth())

class Mux2(Logic):
    def __init__(self, parent, name: str, sel: Wire, sel0: Wire, sel1: Wire, r: Wire):
        super().__init__(parent, name)
        self.sel = self.addIn("sel", sel)
        self.sel0 = self.addIn("sel0", sel0)
        self.sel1 = self.addIn("sel1", sel1)
        self.r = self.addOut("r", r)

    def propagate(self):
        if (self.sel.get() & 1):
            self.r.put(self.sel1.get())
        else:
            self.r.put(self.sel0.get())


class Repeat(Logic):
    def __init__(self, parent, name: str, i: Wire, r: Wire):
        super().__init__(parent, name)

        if (i.getWidth() != 1):
            raise Exception('input wire should be 1 bit')
        self.i = self.addIn('in', i)
        self.r = self.addOut('r', r)

    def propagate(self):
        w = self.r.getWidth()
        wp = (1 << w) - 1
        wf = 0

        v = self.i.get()
        if (v):
            self.r.put(wp)
        else:
            self.r.put(wf)

# deprecated use OneHotMux
class Select(Logic):
    def __init__(self, parent, name: str, sels:list, ins:list, r: Wire):
        super().__init__(parent, name)

        final = []

        for idx, sel in enumerate(sels):
            inv = ins[idx]
            self.addIn('sel{}'.format(idx), sel)
            self.addIn('in{}'.format(idx), inv)
            selx = self.wire('selx{}'.format(idx), inv.getWidth())
            and_sel = self.wire('and_sel{}'.format(idx), inv.getWidth())

            Repeat(self, 'sel{}'.format(idx), sel, selx)
            And2(self, 'and{}'.format(idx), selx, inv, and_sel)
            final.append(and_sel)

        self.addOut('r', r)
        Or(self, 'or', final, r)


class OneHotMux(Logic):
    def __init__(self, parent, name: str, sels:list, ins:list, r: Wire):
        super().__init__(parent, name)

        final = []

        for idx, sel in enumerate(sels):
            inv = ins[idx]
            
            sel_name = sel.name
            in_name = inv.name 
            self.addIn(sel_name, sel)
            self.addIn(in_name, inv)
            selx = self.wire('selx{}'.format(idx), inv.getWidth())
            and_sel = self.wire('and_sel{}'.format(idx), inv.getWidth())

            Repeat(self, 'sel{}'.format(idx), sel, selx)
            And2(self, 'and{}'.format(idx), selx, inv, and_sel)
            final.append(and_sel)

        self.addOut('r', r)
        Or(self, 'or', final, r)


class OneHotDemux(Logic):
    def __init__(self, parent, name: str, sels:list, a:Wire, outs: list):
        super().__init__(parent, name)

        final = []

        a = self.addIn('a', a)

        for idx, sel in enumerate(sels):
            self.addIn('sel{}'.format(idx), sel)
            self.addOut('out{}'.format(idx), outs[idx])
            
            selx = self.wire('selx{}'.format(idx), a.getWidth())

            Repeat(self, 'sel{}'.format(idx), sel, selx)
            And2(self, 'and{}'.format(idx), selx, a, outs[idx])
            
        
class SelectDefault(Logic):
    def __init__(self, parent, name, sels, ins, default, r):
        super().__init__(parent, name)
        
        self.addIn('default', default)
        self.addOut('r', r)
        
        nextoutput = r
        
        for idx, sel in enumerate(sels):
            inv = ins[idx]
            self.addIn('sel{}'.format(idx), sel)
            self.addIn('in{}'.format(idx), inv)
            previnput = self.wire('t{}'.format(idx), r.getWidth())
            Mux2(self, 'mux{}'.format(idx), sel, previnput, inv, nextoutput)
            
            nextoutput = previnput
            
        Buf(self, 'buf', default, previnput)



class Decoder(Logic):
    """
    A decoder
    """

    def __init__(self, parent: Logic, name: str, a: Wire, b):
        """
        Creates a decoder. The outputs are dynamically created

        Parameters
        ----------
        parent : Logic
            Parent circuit.
        name : str
            Name of the instance.
        a : Wire
            Input signal.
        b : List of Wire
            List of the output signals.

        Returns
        -------
        None.

        """
        super().__init__(parent, name)
        a = self.addIn("a", a)

        for i in range(len(b)):
            lb = self.addOut('b{}'.format(i), b[i])

            EqualConstant(self, 'eq{}'.format(i), a, i, lb)


class Minterm(Logic):
    """
    A minterm in a boolean expression
    """

    def __init__(self, parent: Logic, name: str, bits, value: int, r: Wire):
        """

        Parameters
        ----------
        parent : Logic
            DESCRIPTION.
        name : str
            DESCRIPTION.
        bits : List of Wires (least significant order first)
            DESCRIPTION.
        value : int
            DESCRIPTION.
        r : Wire
            DESCRIPTION.

        Returns
        -------
        None.

        """
        super().__init__(parent, name)
        r = self.addOut('r', r)
        parts = []
        lbits = []
        for i in range(len(bits)):
            lbits.append(self.addIn('b{}'.format(i), bits[i]))
            v = (value >> i) & 1
            if (v == 0):
                nbit = self.wire('n{}'.format(i), 1)
                Not(self, 'n{}'.format(i), lbits[i], nbit)
                parts.append(nbit)
            else:
                parts.append(lbits[i])

        And(self, 'prod', parts, r)
        

class SumOfMinterms(Logic):
    def __init__(self, parent, name, a, minterms:list, r):
        super().__init__(parent, name)
        
        self.addIn('a', a)
        self.addOut('r', r)
        
        bits = self.wires('bits', a.getWidth(), 1)
        BitsLSBF(self, 'bits', a, bits)
        
        acum = []
        for idx, minterm in enumerate(minterms):
            v = self.wire('s{}'.format(idx))
            
            Minterm(self, 'minterm{}'.format(idx), bits, minterm, v)
            acum.append(v)
            
        Or(self, 'sum', acum, r)
        
class ConcatenateMSBF(Logic):
    """
    Concatenate wires circuit in MSBF order
    """
    def __init__(self, parent: Logic, name: str, ins:list, r: Wire):
        super().__init__(parent, name)

        ind_w = []
        total_w = 0
        max_w = r.getWidth()
        
        self.ins = []

        for idx, item in enumerate(ins):
            self.ins.append(self.addIn('in{}'.format(idx), item))
            total_w += item.getWidth()
            ind_w.append(item.getWidth())

        if (total_w > max_w):
            raise Exception('Combined input widths {} larger than result width {}>{}'.format(ind_w, total_w, max_w))
            
        self.r = self.addOut('r', r)

    def propagate(self):
        value = 0
        last = 0
        for idx, item in enumerate(self.ins):
            value = value << item.getWidth()
            value = value | item.get()
            
        self.r.put(value)
        
class ConcatenateLSBF(Logic):
    """
    Concatenate wires circuit in MSBF order
    """
    def __init__(self, parent: Logic, name: str, ins:list, r: Wire):
        super().__init__(parent, name)

        total_w = 0
        max_w = r.getWidth()
        
        self.ins = []

        for idx, item in enumerate(ins):
            self.ins.append(self.addIn('in{}'.format(idx), item))
            total_w += item.getWidth()

        if (total_w > max_w):
            raise Exception('Combined input widths larger than result width {}>{}'.format(total_w, max_w))
            
        self.ins.reverse()
        self.r = self.addOut('r', r)

    def propagate(self):
        value = 0
        last = 0

        for idx, item in enumerate(self.ins):
            value = value << item.getWidth()
            value = value | item.get()
            
        self.r.put(value)
        
class Range(Logic):
    def __init__(self, parent: Logic, name: str, a: Wire, high: int, low: int, r: Wire):
        super().__init__(parent, name)

        self.a = self.addIn('a', a)
        self.r = self.addOut('r', r)
        self.high = high
        self.low = low

    def propagate(self):
        value = self.a.get()
        mask = (1 << (self.high - self.low + 1)) - 1
        newvalue = (value >> self.low) & mask
        self.r.put(newvalue)

class Digit7Segment(Logic):
    def __init__(self, parent, name, v, led):
        super().__init__(parent, name)
        
        assert(led.getWidth() == 7)
        
        self.addIn('v', v)
        self.addOut('led', led)
        
        
        a_minterms = [0,2,3,5,6,7,8,9,0xA,0xC,0xE,0xF]
        b_minterms = [0,1,2,3,4,7,8,9,0xA,0xd]
        c_minterms = [0,1,3,4,5,6,7,8,9,0xA,0xb,0xd]
        d_minterms = [0,2,3,5,6,8,0xb,0xC,0xd,0xE]
        e_minterms = [0,2,6,8,0xA,0xb,0xC,0xd,0xE,0xF]
        f_minterms = [0,4,5,6,8,9,0xA,0xb,0xC,0xE,0xF]
        g_minterms = [2,3,4,5,6,8,9,0xA,0xb,0xd,0xE,0xF]
        
        
        a = self.wire('a')
        b = self.wire('b')
        c = self.wire('c')
        d = self.wire('d')
        e = self.wire('e')
        f = self.wire('f')
        g = self.wire('g')
        
        na = self.wire('na')
        nb = self.wire('nb')
        nc = self.wire('nc')
        nd = self.wire('nd')
        ne = self.wire('ne')
        nf = self.wire('nf')
        ng = self.wire('ng')
        
        ConcatenateLSBF(self, 'led', [a,b,c,d,e,f,g], led)
        
        SumOfMinterms(self, 'a', v, a_minterms, a)
        SumOfMinterms(self, 'b', v, b_minterms, b)
        SumOfMinterms(self, 'c', v, c_minterms, c)
        SumOfMinterms(self, 'd', v, d_minterms, d)
        SumOfMinterms(self, 'e', v, e_minterms, e)
        SumOfMinterms(self, 'f', v, f_minterms, f)
        SumOfMinterms(self, 'g', v, g_minterms, g)
        
class PriorityEncoder(Logic):
    def __init__(self, parent, name, a, r, inc_priority=True):
        from ..helper import LogicHelper
        super().__init__(parent, name)
        assert(len(a) == len(r))

        # Make a copy so that reverse does not affect the original list
        a = a.copy()
        r = r.copy()
        
        # It the priority is decreasing, the a[0] is the more priorized
        # Otherwise is the higher index
        if (inc_priority):
            a.reverse()
            r.reverse()
        
        hlp = LogicHelper(self)
        
        last = None
        
        for i in range(len(a)):
            self.addIn('a{}'.format(i), a[i])
            self.addOut('r{}'.format(i), r[i])
            
            if (last is None):
                Buf(self, 'r{}'.format(i), a[i], r[i])
                last = hlp.hw_buf(a[i])
            else:
                And2(self, 'r{}'.format(i), a[i], hlp.hw_not(last), r[i])
                last = hlp.hw_or2(last, a[i])


# -*- coding: utf-8 -*-
"""
Created on Sun Apr  2 13:59:30 2023


Copyright (C) 2023 dcr
Copyright (C) 2023 Victor Suarez Rovere <suarezvictor@gmail.com>
"""

from .. import *
from deprecated import deprecated

import ast
from .astutils import * 


def getBody(node, slist):
    # the Module node contains a body, that contains a list, containting
    # a function definition with a body
    
    # AST visitors can not deal directly with list, we wrap them in 
    # a dummy verilog body object
    var = VerilogWireDeclarations()
    process = VerilogProcess(node.body[0].body, slist)
    return VerilogBody(var, process)
    
class Python2VerilogTranspiler:

    def __init__(self, obj:Logic, methodName:str, slist:str):
        self.obj = obj
        self.methodName = methodName
        self.slist = slist
        self.signals = {}
        self.indent = 0

    def getIndent(self):
        return ' ' * (self.indent * 4)


    def transpileRTL(self):
        '''
        Transpile RTL style behavioural descriptions

        Returns
        -------
        str
            the equivalent RTL

        '''
        module = getMethod(self.obj, self.methodName)
    
        node = getBody(module, self.slist)
    
        assert(isinstance(node, ast.AST))
        
        node = ReplaceIf().visit(node)
        node = ReplaceWireCalls().visit(node)
        node = ReplaceExpr().visit(node)
        node = ReplaceBinOp().visit(node)
        node = ReplaceAttribute().visit(node)
        node = ReplaceConstant().visit(node)
        node = ReplaceAssign().visit(node)
 
        node = FlattenOperators().visit(node)

        return Python2VerilogTranspiler.toVerilog(node)
        #return self.transpileUnknown(node)
        
    @staticmethod
    def toVerilog(node):
        '''
        Tranlates to Verilog a node by calling its toVerilog method
        For lists, in invokes the same method to the elements of the list

        Parameters
        ----------
        node : TYPE
            DESCRIPTION.

        Returns
        -------
        toV : TYPE
            DESCRIPTION.

        '''
        # print('transpiling', type(node))
        if (isinstance(node, ast.Name)): #FIXME: move code
            return node.id

        if (isinstance(node, ast.IfExp)): #FIXME: move code
            test = Python2VerilogTranspiler.toVerilog(node.test)
            body = Python2VerilogTranspiler.toVerilog(node.body)
            orelse = Python2VerilogTranspiler.toVerilog(node.orelse)
            return test + " ? " + body + " : " + orelse;

        if (isinstance(node, list)):
            str = ''
            for item in node:
                str += Python2VerilogTranspiler.toVerilog(item)
            return str
        
        toV = getattr(node, 'toVerilog', None)
        
        if (toV is None):
            return '/*No toVerilog for {}*/'.format(type(node))
        else:
            return toV()
    
    def getExtraDeclarations(self):
        from py4hw.rtl_generation import getValidVerilogName

        str = ""
        
        portNames = []
        
        
        for inp in self.obj.inPorts:
            portNames.append(getValidVerilogName(inp.name))

        for outp in self.obj.outPorts:
            portNames.append(getValidVerilogName(outp.name))        
        
        #print('SIGNALS:', self.signals)
        #print('PORT NAMES:', portNames)

        extra = [x for x in self.signals if x not in portNames]
        
        # @todo we should analyze the number of possible values of 
        # extra signals to decide their width. By now we consider a worst
        # case scenario with extra signals all requiring a maximum of 8 bits
        # this will be simplyfied during synthesis if less bits are required
        # but it will cause a BUG if the required bits are higher
        for sig in extra:
            str += "uint8_t " + sig + " = 0;\n" 
            
        return str
    
    def transpileUnknown(self, line):
        # @deprecated use toVerilog
        if (type(line) == ast.If):
            return self.transpileIf(line);
        if (type(line) == ast.Compare):
            return self.transpileCompare(line)
        if (type(line) == ast.Assign):
            return self.transpileAssign(line)
        if (type(line) == ast.BinOp):
            return self.transpileBinOp(line)
        if (type(line) == ast.Attribute):
            return self.transpileAttribute(line)
        if (isinstance(line, ast.Name)):
            return self.transpileName(line)
        if (type(line) == ast.Eq):
            return "==";
        if (type(line) == ast.Constant):
            return "{}".format(getAstValue(line))
        if (type(line) == ast.Num):
            return "{}".format(int(line.n))
        
        if (type(line) == list):
            str = ""
            for item in line:
                str += self.transpileUnknown(item) 
            return str

        if (type(line) == ast.Expr):
            return self.transpileExpr(line)
        
        else:
            str= "type = {}\n".format(type(line))
            
            str += ast.dump(line) + "\n"
            
            return str
        
    # @deprecated
    def transpileAttribute(self, line:ast.Attribute):
        return line.attr
            
    # @deprecated
    def transpileExpr(self, line:ast.Expr):
        return self.transpileUnknown(line.value)
        
    # @deprecated
    def transpileBinOp(self, line:ast.BinOp):
        return "(" + self.transpileUnknown(line.left) + self.transpileOp(line.op) + self.transpileUnknown(line.right) + ")";

    # @deprecated
    def transpileOp(self, op):
        if (isinstance(op, ast.BitAnd)):
            return '&'
        elif (isinstance(op, ast.BitOr)):
            return '|'
        elif (isinstance(op, ast.BitXor)):
            return '^'
        else:
            raise Exception('Op not supported: {}'.format(op))
        
    # @deprecated
    def transpileAssign(self, line:ast.Assign):
        raise Exception('deprecated')
        # targets = line.targets 
        # if (len(targets) > 1):
        #     return ast.dump(line)
        
        # var = getAstName(targets[0])
        
        # self.signals[var] = var
        
        # return var + " <= " + self.transpileUnknown(line.value) + ";\n" + self.getIndent()  ;
    
    # @deprecated
    def transpileIf(self, line:ast.If):
        str = "if " + self.transpileUnknown(line.test) + "\n"
        str += self.getIndent() + "{\n"
        self.indent += 1
        str += self.getIndent() + self.transpileUnknown(line.body)
        self.indent -= 1
        str +=  "}\n"
        
        if (not(line.orelse is None)):
            str += self.getIndent() + "else \n"
            str += self.getIndent() + "{\n"
            self.indent += 1
            str += self.getIndent() + self.transpileUnknown(line.orelse)
            self.indent -= 1
            str +=  "}\n"

        return str
    
    # @deprecated
    def transpileCompare(self, line:ast.Compare):
        str = "("
        str += self.transpileUnknown(line.left)
        str += self.transpileUnknown(line.ops)
        str += self.transpileUnknown(line.comparators)
        str += ")"
        return str;
    
    def transpileName(self, line:ast.Name):
        str =  getValidVerilogName(line.id)
        self.signals[str] = str
        return str
    
    
    
class ReplaceIf(ast.NodeTransformer):
        
    def visit_If(self, node):
        #print('transpiling if')
        condition = node.test
        
        positive = node.body
        negative = node.orelse
        
        node2 = VerilogIf(condition, positive, negative)

        node2 = ast.NodeTransformer.generic_visit(self, node2)
        return node2
        

class ReplaceWireCalls(ast.NodeTransformer):
        
    def visit_Call(self, node):
        from py4hw.rtl_generation import getAstName

        attr = getAstName(node.func)
        
        #print('checking call', attr)
        if (attr == 'get'):
            #if isinstance(node.func.value, ast.Attribute):
            wirename = getAstName(node.func.value)
            #print('REPLACING GET ', wirename)
            return VerilogWire(wirename)
        
        elif (attr == 'prepare'):
            #print('REPLACE WIRE PUTS FUNC:', attr , node.func.value.attr, node.args)
            left = VerilogWire(node.func.value.attr)
            node = VerilogSynchronousAssignment(left, node.args[0])
        
        elif (attr == 'put'):
            #print('REPLACE WIRE PUTS FUNC:', attr , node.func.value.attr, node.args)
            left = VerilogWire(node.func.value.attr)
            right = ast.NodeTransformer.generic_visit(self, node.args[0])
            node = VerilogAsynchronousAssignment(left, right)
        else:
            print('WARNING: unhandled call {}'.format(attr))
                 
        node = ast.NodeTransformer.generic_visit(self, node)
        
        return node
    
    

class ReplaceBinOp(ast.NodeTransformer):
    
    def visit_BinOp(self, node):
        #print('replacing BinOp')
        node =  VerilogOperator(node.left, node.op, node.right)
        node = ast.NodeTransformer.generic_visit(self, node)
        return node

    def visit_Compare(self, node):
        return VerilogOperator(node.left, node.ops, node.comparators)    
    
class ReplaceExpr(ast.NodeTransformer):
    def visit_Expr(self, node):
        
        return node.value

class ReplaceAttribute(ast.NodeTransformer):
    def visit_Attribute(self, node):
        return VerilogWire(node.attr)
    
class ReplaceConstant(ast.NodeTransformer):
    def visit_Constant(self, node):
        return VerilogConstant(node.value)
    
class ReplaceAssign(ast.NodeTransformer):
    def visit_Assign(self, node):
        if (len(node.targets) > 1):
            raise Exception('{} targets! only 1 is supported'.format(len(node.targets)))
            
        left = node.targets[0]
        right = node.value
        
        return VerilogSynchronousAssignment(left, right)
    
    def visit_AugAssign(self, node):
        left = node.target
        right = node.value
        
        newvalue = VerilogOperator(left, node.op, right)
        return VerilogSynchronousAssignment(left, newvalue)

class FlattenOperators(ast.NodeTransformer):
    # If recursive operators are found they are extracted, new wires
    # are created and the structure is flattened
    ic = -1

    def loop_visit(self, node):
        self.anyChange = True
        
        while (self.anyChange):
            self.anyChange = False
            node = self.visit(node)
            
        return node
            
    def visit_VerilogBody(self, node):
        self.top = node
        return ast.NodeTransformer.generic_visit(self, node)
    
    def newName(self):
        self.ic += 1
        return 'i{}'.format(self.ic)
    
    def visit_VerilogAsynchronousAssignment(self, node):
        self.sync = False
        return ast.NodeTransformer.generic_visit(self, node)

    def visit_VerilogSynchronousAssignment(self, node):
        self.sync = True
        return ast.NodeTransformer.generic_visit(self, node)
        
    def visit_VerilogOperator(self, node):
        if (isinstance(node.left, VerilogOperator)):
            #print('we should extract left operator')
            wn = self.newName()
            vwd = VerilogWireDeclaration(wn)
            vw = VerilogWire(wn)
            self.top.wires.wires.append(vwd)
            
            if (self.sync):
                assign = VerilogSynchronousAssignment(vw, node.left)
            else:
                assign = VerilogAsynchronousAssignment(vw, node.left)
                
            self.top.process.body.append(assign)
            node.left = vw
            self.anyChange = True 
            
        if (isinstance(node.right, VerilogOperator)):
            #print('we should extract right operator')
            wn = self.newName()
            vwd = VerilogWireDeclaration(wn)
            vw = VerilogWire(wn)
            self.top.wires.wires.append(vwd)
            
            if (self.sync):
                assign = VerilogSynchronousAssignment(vw, node.right)
            else:
                assign = VerilogAsynchronousAssignment(vw, node.right)
                
            self.top.process.body.append(assign)
            node.right = vw
            self.anyChange = True
            
        return node

        
class VerilogAsynchronousAssignment(ast.AST):
    def __init__(self, left, right):
        self.left = left
        self.right = right
        self._fields = tuple(['left', 'right'])

    def toVerilog(self):
        str = ''
        
        str += Python2VerilogTranspiler.toVerilog(self.left) + ' = '
        str += Python2VerilogTranspiler.toVerilog(self.right) + '; // async\n'
        return str
        
class VerilogSynchronousAssignment(ast.AST):
    def __init__(self, left, right):
        self.left = left
        self.right = right
        self._fields = tuple(['left', 'right'])

    def toVerilog(self):
        str = ''
        
        str += Python2VerilogTranspiler.toVerilog(self.left) + ' = '
        str += Python2VerilogTranspiler.toVerilog(self.right) + '; // sync\n'
        return str

class VerilogOperator(ast.AST):
    def __init__(self, left, op, right):
        if (isinstance(op, list)):
            # we only support single operators
            assert(len(op) == 1)
            op = op[0]
            
        self.left = left
        self.op = self.getOp(op)
        self.right = right
        self._fields = tuple(['left', 'op', 'right'])

    def getOp(self, operator):
        # translated an AST operator into verilog syntax
        if (isinstance(operator, ast.Add)):
            return '+'
        elif (isinstance(operator, ast.BitAnd)):
            return '&'
        elif (isinstance(operator, ast.BitOr)):
            return '|'
        elif (isinstance(operator, ast.BitXor)):
            return '^'
        elif (isinstance(operator, ast.Eq)):
            return '=='
        elif (isinstance(operator, ast.FloorDiv)):
            return '/'
        elif (isinstance(operator, ast.Mult)):
            return '*'
        elif (isinstance(operator, ast.RShift)):
            return '>>'
        elif (isinstance(operator, ast.Lt)):
            return '<'
        elif (isinstance(operator, ast.GtE)):
            return '>='
        else:
            raise Exception('operator {} not supported'.format(type(operator)))
            
    def toVerilog(self):
        str = ''
        
        str += Python2VerilogTranspiler.toVerilog(self.left) 
        str += self.op
        str += Python2VerilogTranspiler.toVerilog(self.right)
        return str


class VerilogWire(ast.AST):
    '''
    AST node for Verilog Wires
    '''
    def __init__(self, name:str):
        self.name = name
        self._fields = tuple(['name', 'dummy'])

    def toVerilog(self):
        return self.name

class VerilogWireDeclaration(ast.AST):
    '''
    AST node for Verilog Wires
    '''
    def __init__(self, name:str):
        self.name = name
        self._fields = tuple(['name', 'dummy'])

    def toVerilog(self):
        return 'int ' + self.name + ';\n'


class VerilogWireDeclarations(ast.AST):
    # wire declaration section at the beginning of the module
    def __init__(self):
        self.wires = []
        self._fields = tuple(['wires', 'dummy'])
        
    def toVerilog(self):
        str = ''
        
        for w in self.wires:
            str += Python2VerilogTranspiler.toVerilog(w)
            
        return str    

class VerilogProcess(ast.AST):
    def __init__(self, body, sensitivity):
        self.body = body
        self.sensitivity_list = sensitivity
        self._fields = tuple(['body', 'sensitivity_list'])
        
    def toVerilog(self):
        str = 'while(always())\n'.format(self.sensitivity_list)
        str += '{\n'
        
        for st in self.body:
            str += Python2VerilogTranspiler.toVerilog(st) 

        str += '}\n'            
        return str            
    
class VerilogBody(ast.AST):
    '''
    AST node to wrap body
    '''
    def __init__(self, wires, process):
        self.wires = wires
        self.process = process
        self._fields = tuple(['wires', 'process'])
    
    def toVerilog(self):
        str = ''
        
        str += '// variable declaration \n'
        str += Python2VerilogTranspiler.toVerilog(self.wires)

        str += '// process \n'
        str += Python2VerilogTranspiler.toVerilog(self.process)
        return str
            
            
class VerilogIf(ast.AST):
    def __init__(self, condition, positive, negative):
        self.condition = condition
        self.positive = positive
        self.negative = negative
        self._fields = tuple(['condition', 'positive', 'negative'])
        
    def toVerilog(self):
        str = 'if (' + Python2VerilogTranspiler.toVerilog(self.condition) + ')\n'
        str += '{\n'
        str += Python2VerilogTranspiler.toVerilog(self.positive) + '\n'
        if self.negative:
            str += '} else {\n'
            str += Python2VerilogTranspiler.toVerilog(self.negative) + '\n'
        str += '}\n'
        return str
    
class VerilogConstant(ast.AST):
    def __init__(self, value):
        self.value = value
        
    def toVerilog(self):
        return '{}'.format(self.value)